% G = gradientonmesh(X,F,C,smooth_level)
%   Compute gradient of a scalar field on a mesh
%      G   - 3D gradient
%      X,F - Mesh with points X and faces F
%      C   - scalar field on mesh
%      smooth_level - smooth initial data and gradient field (default: 50)
%                     this is to compensate for inherent noise in signal
%
%   Example:
%    gradC = gradientonmesh(X,F,C);
%
%    disp_idx = 1:2:npts1;  % show few arrows
%    trisurf(F, X(:,1), X(:,2), X(:,3), C);  shading interp;  daspect([1 1 1]);  axis on; grid on;  material dull;  xlabel('x'); ylabel('y'); zlabel('z');
%    hold on;  quiver3(X(disp_idx,1),X(disp_idx,2),X(disp_idx,3),gradC(disp_idx,1),gradC(disp_idx,2),gradC(disp_idx,3),1);  hold off;
%
function G = gradientonmesh(X,F,C,smooth_level)

%     A       = adjacency(F);  [~, Dinv] = degree(A);
%     diff    = diffmatrix(C,F);  % directed scalar difference on mesh from node i to j
%     W       = weightedadjacency(X,F);
%     W       = W .* diff;  % use signed signal difference (reorient edge vectors toward gradient)
%     D_w     = spdiags(sum(W,2), 0, speye(size(W,1))); %D = diag(sum(W));
%     G       = ((Dinv*(D_w - W.*A))) * X;
%     return;

    if nargin<4;  smooth_level = 50;  end;  % smooth level

    % Smooth surface function
    if smooth_level > 0
        W = weightedadjacency(X,F); % simple laplacian smoothing
        [~,DWinv] = degree(W);
        for iter = 1:smooth_level
            C = DWinv * W * C;
        end
    end

    % Compute gradient orientation
    dc        = diffmatrix(C,F);         % directed scalar difference on mesh from node i to j
    dxinv     = weightedadjacency(X,F);  % inverse distance of edge
    dcdxinv   = dc .* dxinv;             % use signed signal difference (reorient edge vectors toward gradient)
    D         = degree(dcdxinv);         % to build edges
    [~,Dinv]  = degree(abs(dcdxinv));    % to normalize by sum of absolute weights
    dG        = Dinv*(D-dcdxinv) * X;    % average signed deltaC / deltaX
    mG        = sum(dG.^2,2).^.5;
    mGinv     = mG.^-1;  mGinv(mG==0) = 0;
    dG        = bsxfun(@times, dG, mGinv);  % unit vector, gives orientation

    % Compute gradient magnitude
    dcdxinv2  = dc .* dxinv .* dxinv;    % redivide by edge length, to have unit edges when later multiplying by edges
    D         = degree(dcdxinv2);        % to build edges
    mG        = sum(((D-dcdxinv2) * X).*dG, 2);         % sum of dot products
    z         = sum(abs(dotgradient(X,F,dG).*dxinv),2); % sum of absolute dot product
    zinv      = z.^-1;  zinv(z==0) = 0;  % inverse total weight
    mG        = full(mG .* zinv);        % weighted average
    G         = bsxfun(@times, dG, mG);  % magnitude + direction
    
    % Smooth gradient fields
    if smooth_level > 0
        for iter = 1:smooth_level
            G  = DWinv * W * G;   % Smooth whole field
            mG = DWinv * W * mG;   % Smooth magnitude
        end
        G = bsxfun(@times, dG, mG);   % magnitude + direction
    end
    
end

% A = adjacency(F)
%   Build adjacency matrix from triangle list
%      F - triangle list
%      A - adjacency matrix
%
function A = adjacency(F)

    n = max(F(:));
    
    % remove duplicated edges
    rows = [F(:,1); F(:,1); F(:,2); F(:,2); F(:,3); F(:,3)];
    cols = [F(:,2); F(:,3); F(:,1); F(:,3); F(:,1); F(:,2)];
    [rc,idx] = unique([rows,cols], 'rows','first');

    % fill adjacency matrix
    A = sparse(rc(:,1),rc(:,2),1,n,n);    

end


% D = degree(W)
%   Compute the degree matrix from an adjacency
%   W    - a matrix
%   D    - diagonal degree matrix (each elt is the column sum of W)
%   Dinv - inverse diagonal degree matrix
%
function [D,Dinv] = degree(W)

    n    = size(W,1);
    d    = sum(W,2);
    D    = spdiags(d,    0, speye(n)); %D = diag(sum(W));
    dinv = d.^-1;  dinv(d==0) = 0;     % avoid division by zero
    Dinv = spdiags(dinv, 0, speye(n)); %D = diag(sum(W));
    
end


% D = distancematrix(X,F)
%   Compute distance matrix
%   S - scalar field
%   F - triangulation (n x 3, vertex indices for each triangle)
%   D - distance matrix (n x n)
%       if nodes i and j are connected, W(i,j) gives their distance
%       distance is D(i,j) = dist(X(i,:) - X(j,:)))
%
function D = diffmatrix(S,F)

    % compute weights for all links (euclidean distance)
    dists = [S(F(:,1))-S(F(:,2)); ...
             S(F(:,1))-S(F(:,3)); ...
             S(F(:,2))-S(F(:,1)); ...
             S(F(:,2))-S(F(:,3)); ...
             S(F(:,3))-S(F(:,1)); ...
             S(F(:,3))-S(F(:,2))];
         
    % remove duplicated edges
    rows = [F(:,1); F(:,1); F(:,2); F(:,2); F(:,3); F(:,3)];
    cols = [F(:,2); F(:,3); F(:,1); F(:,3); F(:,1); F(:,2)];
    [rc,idx] = unique([rows,cols], 'rows','first');
    dists = dists(idx);

    % fill adjacency matrix
    D = sparse(rc(:,1), rc(:,2), dists);
    
end


% W = weightedadjacency(X,F)
%   Compute weighted adjacency matrix
%   X - vertex position (n x ndim)
%   F - triangulation (n x 3, vertex indices for each triangle)
%   W - weighted adjacency matrix (n x n)
%       if nodes i and j are connected, W(i,j) gives their affinity
%       affinity is the inverse distance W(i,j) = dist(X(i,:) - X(j,:)))^-1
%
function W = weightedadjacency(X,F)

    % compute weights for all links (euclidean distance)
    weights = [sum((X(F(:,1),:)-X(F(:,2),:)).^2,2); ...
               sum((X(F(:,1),:)-X(F(:,3),:)).^2,2); ...
               sum((X(F(:,2),:)-X(F(:,1),:)).^2,2); ...
               sum((X(F(:,2),:)-X(F(:,3),:)).^2,2); ...
               sum((X(F(:,3),:)-X(F(:,1),:)).^2,2); ...
               sum((X(F(:,3),:)-X(F(:,2),:)).^2,2)].^0.5;
           
    weights = weights.^-1;  % inverse distance

    %eps     = 1e-6; % penalize small distances (avoid division by zero)
    %weights = (weights + eps).^-1;
    
    %sigma = ((max(weights)))/3;
    %weights = exp(-weights.^2 / (2*sigma^2));
    
    % remove duplicated edges
    rows = [F(:,1); F(:,1); F(:,2); F(:,2); F(:,3); F(:,3)];
    cols = [F(:,2); F(:,3); F(:,1); F(:,3); F(:,1); F(:,2)];
    [rc,idx] = unique([rows,cols], 'rows','first');
    weights = weights(idx);

    % fill adjacency matrix
    W = sparse(rc(:,1), rc(:,2), weights);
    
end


% d = dotgradient(X,F)
%   Compute dot product between edge and unit gradient, to find gradient norm
%   X - vertex position (n x ndim)
%   F - triangulation (n x 3, vertex indices for each triangle)
%   G - unit gradient vectors, gives orientation
%
%   Eventually, gradient magnitude is weighted sum of dot product
%
function d = dotgradient(X,F,G)

%     dotgr   = [ sum( (X(F(:,1),:)-X(F(:,2),:)) .* G(F(:,1),:) ,2); ...
%                 sum( (X(F(:,1),:)-X(F(:,3),:)) .* G(F(:,1),:) ,2); ...
%                 sum( (X(F(:,2),:)-X(F(:,1),:)) .* G(F(:,2),:) ,2); ...
%                 sum( (X(F(:,2),:)-X(F(:,3),:)) .* G(F(:,2),:) ,2); ...
%                 sum( (X(F(:,3),:)-X(F(:,1),:)) .* G(F(:,3),:) ,2); ...
%                 sum( (X(F(:,3),:)-X(F(:,2),:)) .* G(F(:,3),:) ,2) ];
            
    Xf = [  X(F(:,1),:)-X(F(:,2),:); ...
            X(F(:,1),:)-X(F(:,3),:); ...
            X(F(:,2),:)-X(F(:,1),:); ...
            X(F(:,2),:)-X(F(:,3),:); ...
            X(F(:,3),:)-X(F(:,1),:); ...
            X(F(:,3),:)-X(F(:,2),:)  ];
        
    Gf = [  G(F(:,1),:); ...
            G(F(:,1),:); ...
            G(F(:,2),:); ...
            G(F(:,2),:); ...
            G(F(:,3),:); ...
            G(F(:,3),:)  ];
        
    dotgr = dot(Xf', Gf')';
           
    % remove duplicated edges
    rows = [F(:,1); F(:,1); F(:,2); F(:,2); F(:,3); F(:,3)];
    cols = [F(:,2); F(:,3); F(:,1); F(:,3); F(:,1); F(:,2)];
    [rc,idx] = unique([rows,cols], 'rows','first');
    dotgr = dotgr(idx);

    % fill adjacency matrix
    d = sparse(rc(:,1), rc(:,2), dotgr);
    
end



