% [corr, M1, M2] = spectralign(M1, M2 opts)
%   Spectral Alignment of Brain Surfaces
%
%   M1, M2 - two meshes with fields:
%            .X - mesh vertices
%            .F - mesh faces
%            .S - mesh surface data (e.g., sulcal depth)
%            .P - mesh parcel id (optional, to monitor Dice scores)
%   opts   - Algorithm options:
%            .nbmodes - nb of eigenmodes to use
%            .wS - weight of surface data (default: 0.5)
%            .wdS - weight of surface data gradients (default: 0.8)
%            .min_err - minimal tolerated error, zero to disable (default: 1)
%            .demons.active - register with final surface demons (default: 1)
%            .demons.sigma_x - regularization factor (default: 1.0)
%            .demons.smooth_level - regularization bandwidth (default: 30)
%            .demons.niter - nb of itertions (default: 10)
%
%   corr   - correspondence maps
%            .corr12 - map from mesh 1 to mesh 2
%            .corr21 - map from mesh 2 to mesh 1
%            .R12    - spectral transformation from embedding 1 to 2
%            .R21    - spectral transformation from embedding 2 to 1
%
%   M1, M2 - two warped meshes as output, with eigen components:
%            .E.U - spectral eigenvectors
%            .E.V - spectral eigenvalues
%            .E.X - spectral coordinates (D^-.5 * U * V^-.5)
%
% References:
%
% [1] Brain Transfer: Spectral Analysis of Cortical Surfaces and Functional Maps
%     Lombaert, Arcaro, Ayache, IPMI, 2015
%
% [2] -, 2018
%
% herve.lombaert@gmail.com, January 2018
%
function [corr, M1, M2, X1p, X2p] = spectralign(M1, M2, opts)

    do_display(0);


    %% Default parameters
    if nargin<3;  opts = struct();  end;
    
    if ~isfield(opts, 'nbmodes');  opts.nbmodes   = 65;                       end;  % nb of eigen components
    if ~isfield(opts, 'wS');       opts.wS        = 0.5;                      end;  % weight of surface data
    if ~isfield(opts, 'wdS');      opts.wdS       = 0.8;                      end;  % weight of surface data gradients
    if ~isfield(opts, 'kr');       opts.kr        = {1:4, 1:10, 1:35, 1:65};  end;  % nb eigenvectors for reconstruction of velocities
    if ~isfield(opts, 'niter');    opts.niter     = {20,  15,   5,    5   };  end;  % nb iterations per reconstruction of velocities
    if ~isfield(opts, 'min_err');  opts.min_err   = 1;                        end;  % minimal tolerated error

    if ~isfield(opts, 'demons');   opts.demons    = struct();  end;  % do demons or not
    
    if ~isfield(opts.demons, 'active');        opts.demons.active       = 1;   end;  % update step
    if ~isfield(opts.demons, 'sigma_x');       opts.demons.sigma_x      = 1;   end;  % regularization factor
    if ~isfield(opts.demons, 'smooth_level');  opts.demons.smooth_level = 30;  end;  % regularization bandwidth
    if ~isfield(opts.demons, 'step');          opts.demons.step         = 1;   end;  % regularization factor
    if ~isfield(opts.demons, 'niter');         opts.demons.niter        = 10;  end;  % nb of iterations
    if ~isfield(opts.demons, 'do12');          opts.demons.do12         = 1;   end;  % register 1 to 2
    if ~isfield(opts.demons, 'do21');          opts.demons.do21         = 1;   end;  % register 2 to 1
    
    if isfield(opts,'corr');  corr = opts.corr;  end;
    
    mytic('SpectrAligner');

    %% Computing embeddings
    mytic('Eigendecomposition');
    M1.E = embedding(M1,opts.nbmodes,struct('mode','randomwalk'));
    M2.E = embedding(M2,opts.nbmodes,struct('mode','randomwalk'));
    M1.E.Upinv = pinv(M1.E.U);
    M2.E.Upinv = pinv(M2.E.U);
    mytoc;

    X1 = M1.E.X(:,2:4);
    F1 = M1.F;

    X2 = M2.E.X(:,2:4);
    F2 = M2.F;

    dx_X1 = (sum(M1.E.V(2:4).^-.5)/3);  % avg edge length on embedding
    dx_X2 = (sum(M2.E.V(2:4).^-.5)/3);  % avg edge length on embedding
    
    %% Predictor functions
    if isfield(M1,'S') && isfield(M2,'S')
        
        %% Gradient on meshes
        C1  = M1.S;
        dC1 = gradientonmesh(M1.E.X(:,2:4), M1.F, M1.S, 50);
        dC1 = dC1 / dx_X1;

        C2  = M2.S;
        dC2 = gradientonmesh(M2.E.X(:,2:4), M2.F, M2.S, 50);
        dC2 = dC2 / dx_X2;
        
        %% Build data term
        D1 = [  opts.wS*C1  opts.wdS*dC1  ];
        D2 = [  opts.wS*C2  opts.wdS*dC2  ];

    else
        
        %% Wave Kernel Signature
        C1 = waveKernelSignature(M1.E.U, M1.E.V, M1.E.W, 4);
        C2 = waveKernelSignature(M2.E.U, M2.E.V, M2.E.W, 4);

        C1   = C1 * dx_X1;
        C2   = C2 * dx_X2;
        
        %% Build data term
        D1 = [  opts.wS*C1  ];
        D2 = [  opts.wS*C2  ];

    end

    if do_display
        [ls,ld] = isolines(X1*1.01,F1,C1(:,1),[Inf +.5],'k',3);  
    end

    % Nearest Neighbors on Extended Embedding
    Z1    = [ X1  D1 ];
    Z2    = [ X2  D2 ];
    Z1o   = Z1;
    Z2o   = Z2;

    if do_display
        subplot(211);  displaymesh(X1*.99,F1,C1(:,1));  caxis([-1 1] * 1.5);
        subplot(212);  displaymesh(X2*.99,F2,C2(:,1));  caxis([-1 1] * 1.5);
        drawnow;
    end

    %
    mytic('Align Embeddings');
    do_verbose(0);
    last_errZ = 1e10;
    last_Z1   = Z1;
    last_Z2   = Z2;
    for iter_recon = 1:numel(opts.niter)
        for iter = 1:opts.niter{iter_recon}
            %% Match extended embeddings
            corr.corr12 = match( Z1o, Z2 );  % corr12 closest pts of M1 in M2
            corr.corr21 = match( Z2o, Z1 );  % corr21 closest pts of M2 in M1
            C12 = sparse(1:M1.n, corr.corr12, 1, M1.n, M2.n);
            C21 = sparse(1:M2.n, corr.corr21, 1, M2.n, M1.n);

            %% Monitor Errors
            errZ  = sum( sum( (Z1(corr.corr21,:)   - Z2o).^2, 2) );        % total error
            %errX  = sum( sum( (Z1(corr.corr21,1:3) - Z2o(:,1:3)).^2, 2) ); % embedding error 
            %errC  = sum( sum( (C1(corr.corr21,:)   - C2).^2, 2) );         % sulcal depth error
            %errdC = sum( sum( (dC1(corr.corr21,:)  - dC2).^2, 2) );        % gradient sulcal depth error
            %disp(['(iter ' num2str(iter) ' - ' num2str(max(opts.kr{iter_recon})) ' bases) Err(Z): ' num2str(errZ,'%.2f') ' = Err(X): ' num2str(errX,'%.2f') ' + ' num2str(opts.wS) '^2*Err(C): ' num2str(errC,'%.2f') ' + ' num2str(opts.wdS) '^2*Err(dC): ' num2str(errdC,'%.2f') ' -- Acc: ' num2str(acc*100,'%.1f') '%']);

            % Monitor Dice
            if isfield(M1,'Pm') & isfield(M2,'Pm')
                acc   = sum((M1.Pm(corr.corr21) == M2.Pm)) / M2.n;             % accuracy
                [s,labels] = dice(M1.Pm(corr.corr21),M2.Pm);                   % dice Mn
                disp(['(iter ' num2str(iter) ' - ' num2str(max(opts.kr{iter_recon})) ' bases) Dice Mn (avgstd [min-max]): ' num2str(100*mean(s),'%.1f') '' num2str(100*std(s),'%.1f') '% [' num2str(100*min(s),'%.1f') '-' num2str(100*max(s),'%.1f') '] -- Err(Z): ' num2str(errZ,'%.2f') ' -- Acc: ' num2str(acc*100,'%.1f') '%']);
                %disp(['(iter ' num2str(iter) ' - ' num2str(max(opts.kr{iter_recon})) ' bases) Dice Mn (avgstd [min-max]): ' num2str(100*mean(s),'%.1f') '' num2str(100*std(s),'%.1f') '% [' num2str(100*min(s),'%.1f') '-' num2str(100*max(s),'%.1f') '] -- Err(Z): ' num2str(errZ,'%.2f') ' = Err(X): ' num2str(errX,'%.2f') ' + ' num2str(opts.wS) '^2*Err(C): ' num2str(errC,'%.2f') ' + ' num2str(opts.wdS) '^2*Err(dC): ' num2str(errdC,'%.2f') ' -- Acc: ' num2str(acc*100,'%.1f') '%']);
            elseif isfield(M1,'P') & isfield(M2,'P')
                acc   = sum((M1.P(corr.corr21) == M2.P)) / M2.n;               % accuracy
                [s,labels] = dice(M1.P(corr.corr21),M2.P);                     % dice FS
                disp(['(iter ' num2str(iter) ' - ' num2str(max(opts.kr{iter_recon})) ' bases) Dice FS (avgstd [min-max]): ' num2str(100*mean(s),'%.1f') '' num2str(100*std(s),'%.1f') '% [' num2str(100*min(s),'%.1f') '-' num2str(100*max(s),'%.1f') '] -- Err(Z): ' num2str(errZ,'%.2f') ' -- Acc: ' num2str(acc*100,'%.1f') '%']);
                %disp(['(iter ' num2str(iter) ' - ' num2str(max(opts.kr{iter_recon})) ' bases) Dice FS (avgstd [min-max]): ' num2str(100*mean(s),'%.1f') '' num2str(100*std(s),'%.1f') '% [' num2str(100*min(s),'%.1f') '-' num2str(100*max(s),'%.1f') '] -- Err(Z): ' num2str(errZ,'%.2f') ' = Err(X): ' num2str(errX,'%.2f') ' + ' num2str(opts.wS) '^2*Err(C): ' num2str(errC,'%.2f') ' + ' num2str(opts.wdS) '^2*Err(dC): ' num2str(errdC,'%.2f') ' -- Acc: ' num2str(acc*100,'%.1f') '%']);
            end

            % Stop if error grows
            if (errZ > last_errZ)
                Z1 = last_Z1;
                Z2 = last_Z2;
                break;
            end
            if (last_errZ - errZ < opts.min_err)
                last_errZ = errZ;
                last_Z1   = Z1;
                last_Z2   = Z2;
                break;
            end
            last_errZ = errZ;
            last_Z1   = Z1;
            last_Z2   = Z2;

            %% Reconstruct embedding
            R21 = M1.E.Upinv(opts.kr{iter_recon},:) * C12 * M2.E.Upinv(opts.kr{iter_recon},:)';  %[u,s,v] = svd(R21m1);  R21m1 = u*v';
            w1  = M1.E.Upinv(opts.kr{iter_recon},:) * M1.E.X(:,opts.kr{iter_recon});
            w2  = M2.E.Upinv(opts.kr{iter_recon},:) * M2.E.X(:,opts.kr{iter_recon});
            w   = R21 * w2;
            Y1  = M1.E.U(:,opts.kr{iter_recon}) * w;  % reconstruct embedding
            Y1  = Y1(:,2:4);
            Z1  = [ Y1(:,1:3)  D1 ];  % updated embedding

            %% Reconstruct embedding
            R12 = M2.E.Upinv(opts.kr{iter_recon},:) * C21 * M1.E.Upinv(opts.kr{iter_recon},:)';  %[u,s,v] = svd(R21m1);  R21m1 = u*v';
            w   = R12 * w1;
            Y2  = M2.E.U(:,opts.kr{iter_recon}) * w;  % reconstruct embedding
            Y2  = Y2(:,2:4);
            Z2  = [ Y2(:,1:3)  D2 ];  % updated embedding

            %% Smooth Deformation (this moves points closer to their destination, speeding up knn)
            dX1 = Z2o(corr.corr12,1:3) - Z1(:,1:3);  % raw update
            dX1 = M1.E.U(:,opts.kr{iter_recon}) * (M1.E.Upinv(opts.kr{iter_recon},:) * dX1);  % smoothed update
            Z1  = [ Z1(:,1:3)+dX1  D1 ];  % updated embedding

            dX2 = Z1o(corr.corr21,1:3) - Z2(:,1:3);  % raw update
            dX2 = M2.E.U(:,opts.kr{iter_recon}) * (M2.E.Upinv(opts.kr{iter_recon},:) * dX2);  % smoothed update
            Z2  = [ Z2(:,1:3)+dX2  D2 ];  % updated embedding

    %         %% Update sulcal depth gradients (this doesn't help much, embeddings are already smooth, terribly slows down iterations)
    %         dC1 = gradientonmesh(Z1(:,1:3),F1,C1);
    %         dC2 = gradientonmesh(Z2(:,1:3),F2,C2);
    %         Z1  = [ Z1(:,1:3)  opts.wS*C1  opts.wdS*dC1 ];
    %         Z2  = [ Z2(:,1:3)  opts.wS*C2  opts.wdS*dC2 ];

            % Display
            if do_display
                %%
                subplot(211);
                idx = 1:25:M1.n;  % show few arrows
                displaymesh(X1*.99,F1,C1(:,1));  caxis([-1 1] * 1.5);

                subplot(212);
                idx = 1:25:M2.n;  % show few arrows
                x2  = Z2(:,1:3);  % updated embedding
                displaymesh(x2*.99,F2,C2(:,1));  caxis([-1 1] * 1.5);
    %             hold on;  quiver3(x2(idx,1),x2(idx,2),x2(idx,3), dX2(idx,1),dX2(idx,2),dX2(idx,3), 2);  hold off;

                view(0,-90);
                link_subplots; link_light

                drawnow;
            end
        end

        last_errZ = 1e10;
        last_Z1   = Z1;
        last_Z2   = Z2;

    end
    do_verbose(1);
    mytoc;
    do_verbose(0);

    % Correspondence on embeddings
    corr12o = corr.corr12;
    corr21o = corr.corr21;
    do_verbose(0);
    corr.corr12 = match( Z1(:,1:3), Z2o(:,1:3) );  % corr12 closest pts of M1 in M2
    corr.corr21 = match( Z2(:,1:3), Z1o(:,1:3) );  % corr21 closest pts of M2 in M1
    do_verbose(1);

    % Update Transformation
    C12 = sparse(1:M1.n, corr.corr12, 1, M1.n, M2.n);
    C21 = sparse(1:M2.n, corr.corr21, 1, M2.n, M1.n);
    R12 = M1.E.Upinv(1:opts.nbmodes,:) * C12 * M2.E.Upinv(1:opts.nbmodes,:)';  %[u,s,v] = svd(R21m1);  R21m1 = u*v';
    R21 = M1.E.Upinv(1:opts.nbmodes,:) * C12 * M2.E.Upinv(1:opts.nbmodes,:)';  %[u,s,v] = svd(R21m1);  R21m1 = u*v';

    corr.R12 = R12;
    corr.R21 = R21;
    
    % Do final Spectral Demons
    if opts.demons.active
        
        %% Demons
        X1o = Z1o(:,1:3);  % Z1o is original Z1
        X2o = Z2o(:,1:3);  % Z2o is original Z2
        X1  = Z1(:,1:3);   % Z1 should look like Z2o
        X2  = Z2(:,1:3);   % Z2 should look like Z1o

        C1s  = C1(:,1);
        C2s  = C2(:,1);
        dC1s = gradientonmesh(X1,F1,C1s,10);
        dC2s = gradientonmesh(X2,F2,C2s,10);

        idx = 1:25:M2.n;  % show few arrows

        % Parameters
        do_display(0);

        L1 = M1.E.Dinv * M1.E.W; % smooth filter
        L2 = M2.E.Dinv * M2.E.W;

        do_verbose(0);
        mytic('Spectral Demons');
        for iter = 1:opts.demons.niter

            if opts.demons.do12;  corr12d = match( X1, X2o );  end;
            if opts.demons.do21;  corr21d = match( X2, X1o );  end;

            %% Monitor Errors
            if opts.demons.do21
                errX  = sum( sum( (X1(corr21d,1:3) - X2o).^2, 2) );                     % embedding error 
                errC  = sum( sum( (C1s(corr21d,:)  - C2s).^2, 2) );                     % sulcal depth error
                %errdC = sum( sum( (dC1s(corr21d,:)/dx_X1 - dC2s/dx_X2).^2, 2) );        % gradient sulcal depth error
                %errZ  = errX + opts.wS^2*errC + opts.wdS^2*errdC;                       % total error
                errZ  = errX + opts.wS^2*errC;                                          % total error
                %disp(['(iter ' num2str(iter) ') Err(Z): ' num2str(errZ,'%.2f') ' = Err(X): ' num2str(errX,'%.2f') ' + ' num2str(opts.wS) '^2*Err(C): ' num2str(errC,'%.2f') ' + ' num2str(opts.wdS) '^2*Err(dC): ' num2str(errdC,'%.2f') ' -- Acc: ' num2str(acc*100,'%.1f') '%']);

                % Monitor Dice
                if isfield(M1,'Pm') && isfield(M2,'Pm')
                    acc   = sum((M1.Pm(corr21d) == M2.Pm)) / M2.n;                      % accuracy
                    [s,labels] = dice(M1.Pm(corr21d), M2.Pm);                           % dice Mn
                    disp(['(iter ' num2str(iter) ') Dice Mn (avgstd [min-max]): ' num2str(100*mean(s),'%.1f') '' num2str(100*std(s),'%.1f') '% [' num2str(100*min(s),'%.1f') '-' num2str(100*max(s),'%.1f') '] -- Err(Z): ' num2str(errZ,'%.2f') ' = Err(X): ' num2str(errX,'%.2f') ' + ' num2str(opts.wS) '^2*Err(C): ' num2str(errC,'%.2f') ' -- Acc: ' num2str(acc*100,'%.1f') '%']);
                    %disp(['(iter ' num2str(iter) ') Dice Mn (avgstd [min-max]): ' num2str(100*mean(s),'%.1f') '' num2str(100*std(s),'%.1f') '% [' num2str(100*min(s),'%.1f') '-' num2str(100*max(s),'%.1f') '] -- Err(Z): ' num2str(errZ,'%.2f') ' = Err(X): ' num2str(errX,'%.2f') ' + ' num2str(opts.wS) '^2*Err(C): ' num2str(errC,'%.2f') ' + ' num2str(opts.wdS) '^2*Err(dC): ' num2str(errdC,'%.2f') ' -- Acc: ' num2str(acc*100,'%.1f') '%']);
                elseif isfield(M1,'P') && isfield(M2,'P')
                    acc   = sum((M1.P(corr21d) == M2.P)) / M2.n;                        % accuracy
                    [s,labels] = dice(M1.P(corr21d), M2.P);                             % dice FS
                    disp(['(iter ' num2str(iter) ') Dice FS (avgstd [min-max]): ' num2str(100*mean(s),'%.1f') '' num2str(100*std(s),'%.1f') '% [' num2str(100*min(s),'%.1f') '-' num2str(100*max(s),'%.1f') '] -- Err(Z): ' num2str(errZ,'%.2f') ' = Err(X): ' num2str(errX,'%.2f') ' + ' num2str(opts.wS) '^2*Err(C): ' num2str(errC,'%.2f') ' -- Acc: ' num2str(acc*100,'%.1f') '%']);
                    %disp(['(iter ' num2str(iter) ') Dice FS (avgstd [min-max]): ' num2str(100*mean(s),'%.1f') '' num2str(100*std(s),'%.1f') '% [' num2str(100*min(s),'%.1f') '-' num2str(100*max(s),'%.1f') '] -- Err(Z): ' num2str(errZ,'%.2f') ' = Err(X): ' num2str(errX,'%.2f') ' + ' num2str(opts.wS) '^2*Err(C): ' num2str(errC,'%.2f') ' + ' num2str(opts.wdS) '^2*Err(dC): ' num2str(errdC,'%.2f') ' -- Acc: ' num2str(acc*100,'%.1f') '%']);
                end
            end

            % Update Embedding 1
            if opts.demons.do12
                normg2 = sum((.5*dC1s + .5*dC2s(corr12d,:)).^2,2);
                diff   = (C2s(corr12d,:) - C1s) * dx_X1;
                scale  = - diff ./ (normg2 + diff.^2 * opts.demons.sigma_x^2) / dx_X1;  % levenberg-marquart descent
                scale( (normg2+diff.^2) < 1e-6 ) = 0;                                   % vanishing gradients
                dX1    = bsxfun(@times, (.5*dC1s + .5*dC2s(corr12d,:)), scale);         % symmetric
                for iter_smooth = 1:opts.demons.smooth_level;  dX1 = L1 * dX1;  end;    % Smooth update
                X1     = X1 + opts.demons.step*dX1;                                     % update embedding
            end

            % Update Embedding 2
            if opts.demons.do21
                normg2 = sum((.5*dC2s + .5*dC1s(corr21d,:)).^2,2);
                diff   = (C1s(corr21d,:) - C2s) * dx_X2;
                scale  = - diff ./ (normg2 + diff.^2 * opts.demons.sigma_x^2) / dx_X2;  % levenberg-marquart descent
                scale( (normg2+diff.^2) < 1e-6 ) = 0;                                   % vanishing gradients
                dX2    = bsxfun(@times, (.5*dC2s + .5*dC1s(corr21d,:)), scale);         % symmetric
                for iter_smooth = 1:opts.demons.smooth_level;  dX2 = L2 * dX2;  end;    % Smooth update
                X2     = X2 + opts.demons.step*dX2;                                     % update embedding
            end

            % dC1s = gradientonmesh(X1,F1,C1s);  % so similar, doesn't improve at all, and terribly slow downs iterations
            % dC2s = gradientonmesh(X2,F2,C2s);

            % Display
            if do_display
                displaymesh(X2*.99, M2.F, C2s);  caxis([-1 1] * 1.5);
                view(0,-90); link_light;
            %     hold on;  quiver3(X2(idx,1),X2(idx,2),X2(idx,3), dX2(idx,1),dX2(idx,2),dX2(idx,3), 2);  hold off;
            %     hold on;  isolines(ls,ld, 'k');  hold off;
            %     hold on;  isolines(X1*1.01 ,F1 ,C1, [Inf +.35], 'k', 3);  hold off;
            %     hold on;  isolines(X2*1.01 ,F2 ,C2, [Inf +.35], 'g', 3);  hold off;
                drawnow;
            end
        end
        do_verbose(1);
        mytoc; % Demons
        do_verbose(0);

        corr.corr12 = corr12d;
        corr.corr21 = corr21d;

    end

    % Update Transformation
    C12 = sparse(1:M1.n, corr.corr12, 1, M1.n, M2.n);
    C21 = sparse(1:M2.n, corr.corr21, 1, M2.n, M1.n);
    R12 = M1.E.Upinv(1:opts.nbmodes,:) * C12 * M2.E.Upinv(1:opts.nbmodes,:)';  %[u,s,v] = svd(R21m1);  R21m1 = u*v';
    R21 = M1.E.Upinv(1:opts.nbmodes,:) * C12 * M2.E.Upinv(1:opts.nbmodes,:)';  %[u,s,v] = svd(R21m1);  R21m1 = u*v';

    corr.C12 = C12;
    corr.C21 = C21;
    corr.R12 = R12;
    corr.R21 = R21;
    
    X1p = X1;
    X2p = X2;

    %
    do_verbose(1);
    mytoc; % Spectral Aligner
    do_verbose(0);

    %
    if do_display
        %%
        X1 = Z1o(:,1:3);  C1 = D1(:,1);
        X2 = Z2(:,1:3);   C2 = D2(:,1);
        [ls,ld] = isolines(X1*1.01 ,M1.F ,C1, [Inf +.5], 'k', 3);  
        subplot(211); displaymesh(X1*.99, M1.F, M1.P);  hold on; isolines(ls,ld,'k'); hold off;
        subplot(212); displaymesh(X2*.99, M2.F, M2.P);  hold on; isolines(ls,ld,'k'); hold off;
        view(0,-90); link_subplots; link_light
        drawnow;
    end

end
